(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: code.ml,v 1.16 2007/04/26 17:18:01 weis Exp $ *)

(* The code we generate. *)

open Parsetree;;
open Longident;;
open Asttypes;;
open Printf;;

type lident = string;;
(** Lowercase ident. *)
type uident = string;;
(** Uppercase ident. *)
type iident = string;;
(** Infix ident. *)
type ident = string;;
(** Any kind of ident. *)

type exp =
   | Id of ident
   | Match of exp * clause list
   | Try of exp * clause list
   | Constr of uident * exp list
   | App of exp * exp list
   | AppI of exp * iident * exp
   | If of exp * exp * exp
   | Let of lident * lident list * exp * exp
   | Letrec of lident * lident list * exp * exp
   | Record of (lident * exp) list
   | Pair of exp * exp
   | Ocaml_pattern of Parsetree.pattern
   | Ocaml_expression of Parsetree.expression
   | Function of clause list
   | Comment of string

and clause =
  | Clause of exp * exp
  | Clause_when of exp * exp * exp;;

type structure_item =
   (** Defines a set of possibly recursive functional values *)
   | Pstr_value of Asttypes.rec_flag * (lident * lident list * exp) list
   (** Defines one primitive *)
   | Pstr_primitive of string * Parsetree.value_description
   (** Defines one type *)
   | Pstr_type of (lident * Parsetree.type_declaration) list;;

type structure = structure_item list;;

let topletrec def = Pstr_value (Recursive, def);;

(* Globally predefined identifiers. *)

let underscore = Id "_";;
let x = Id "x"
and y = Id "y"
and z = Id "z"
and t = Id "t"
and u = Id "u";;

(* [genr_args root n] returns the list of strings [root1; ...; rootn].

  Beware: do not modify the code to generate a mere (admittedly simpler) "x",
  since arguments generated by [genr_args] should not be "x" or "u" or "t" or
  "y" or "z": those variables are already used in the generating functions
  as ``global'' (i.e. toplevel) arguments of generated functions.
  If genr_args would produce "x" for instance, we would have a name
  capture between a pattern of a clause and the argument of the
  generated functions.
  To prevent this phenomenon, genr_args always adds an integer after the root
  prefix. Be cautious to preserve this invariant! *)
let genr_args root =
  let rec loop accu = function
    | 0 -> accu
    | n -> loop ((root ^ string_of_int n) :: accu) (n - 1) in
  loop [];;

let genr_fun_args n =
  match genr_args "x" n with
  | [] -> assert false
  | x :: l -> x, l;;

let exps_of_idents = List.map (fun x -> Id x);;

let idents_of_exps =
  let id_of_exp = function
    | Id x -> x
    | _ -> assert false in
  List.map id_of_exp;;

let genr_idargs root n = exps_of_idents (genr_args root n);;

(* [genr_numbered_idargs root n]
   returns the list of strings [1, Id root1; ...; n, Id rootn].
   Beware: same as the one for [genr_args]. *)
let genr_numbered_idargs root n =
  let rec loop accu = function
    | 0 -> accu
    | n -> loop ((n, Id (root ^ string_of_int n)) :: accu) (n - 1) in
  loop [] n;;

(* Application macros. *)

let app1 f x = App (Id f, [x]);; 
let app2 f x y = App (Id f, [x; y]);; 
let app3 f x y z = App (Id f, [x; y; z]);; 
let appi x f y = AppI (x, f, y);;
let constr1 c x = Constr (c, [x]);;
let constr2 c x y = Constr (c, [x; y]);;

(* Other useful auxilliaries. *)

let let0 x v b = Let (x, [], v, b);;
let let1 f x v b = Let (f, [x], v, b);;
let pair x y = Pair (x,y);;
let raise_error s = app1 "raise" (Id s);;

(* Printing code. *)

let rec iter f b = function
  | [] -> ()
  | x :: l -> bprintf b "%a%a" f x (iter f) l;;

let ident b = bprintf b " %s";;

let idents = iter ident;;

let rec exp b = function
  | Id s -> bprintf b "%s" s
  | Ocaml_pattern p -> Pr_otype.pattern b p
  | Ocaml_expression e -> Pr_otype.expression b e
  | Match (e, l) -> bprintf b "\n(match %a with%a)\n" exp e clauses l
  | Try (e, l) -> bprintf b "\n(try %a with%a)\n" exp e clauses l
  | Constr (s, l) -> bprintf b "(%s%a)" s tuple l
  | App (e, l) -> app " " b e l
  | AppI (e1, s, e2) -> bprintf b "(%a %s %a)" exp e1 s exp e2
  | If (e1, e2, e3) ->
    bprintf b "if %a\nthen %a\nelse %a\n" exp e1 exp e2 exp e3
  | Let (fname, args, e1, e2) ->
    bprintf b "\nlet %s%a = %a in\n%a" fname idents args exp e1 exp e2 
  | Letrec (fname, args, e1, e2) ->
    bprintf b "\nlet rec %s%a = %a in\n%a" fname idents args exp e1 exp e2 
  | Record l -> bprintf b "{ %a}" fields l
  | Pair (e1, e2) -> bprintf b "(%a, %a)" exp e1 exp e2
  | Function l -> bprintf b "function\n(%a)\n" clauses l
  | Comment s -> bprintf b "\n(* %s *)\n" s

and sep_exp sep b = bprintf b "%s%a" sep exp

and sep_exps sep = iter (sep_exp sep)

and app sep b e l = bprintf b "(%a%a)" exp e (sep_exps sep) l

and tuple b = function
  | [] -> ()
  | e :: l -> bprintf b "(%a%a)" exp e (sep_exps ", ") l

and clause b = function
  | Clause (e1, e2) -> bprintf b "\n| %a -> %a" exp e1 exp e2
  | Clause_when (e1, w, e2) ->
    bprintf b "\n| %a when %a -> %a" exp e1 exp w exp e2

and clauses b = iter clause b

and field b (name, e) = bprintf b "%s = %a; " name exp e

and fields b = iter field b;;

let rec structure_item b = function
  | Pstr_value (_, []) -> ()
  | Pstr_value (Recursive, ((f, args, e) :: rest)) ->
    bprintf b "\nlet rec %s%a = %a%a;;\n" f idents args exp e print_and rest
  | Pstr_value (Nonrecursive, ((f, args, e) :: rest)) ->
    bprintf b "\nlet %s%a = %a%a;;\n" f idents args exp e print_and rest
  | Pstr_primitive (s, valdesc) ->
    bprintf b "\nexternal %s%a;;\n" s Pr_otype.value_description valdesc
  | Pstr_type tydecls ->
    bprintf b "\ntype %a;;\n" Pr_otype.pstr_type tydecls
  | _ -> failwith "Not yet implemented"

and print_and b =
  List.iter
    (function (f, args, e) ->
     bprintf b "\nand %s%a = %a" f idents args exp e);;

let structure = iter structure_item;;
